#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include "vector.h"

#define R          64
#define R_2        128
#define R_BITS     6
#define R2         128
#define R3         96
#define EDGE       16
#define HEIGHT     20
#define MAX_X      319
#define MAX_Y      199
#define TEX_SIZE   64
#define TEX_BITS   6
#define TEX_MASK   63

char far *light_map;
char far *texture;

const double C = M_PI / 180;

class Light {
  public:
    Vector org;
    Light() { org = 0; }
    Light(Vector org) { Light::org = org; }
} light(Vector(MAX_X / 2, MAX_Y / 2, HEIGHT));

void SetMode(int mode)
{
  asm
  {
      mov    ax,  mode
      int    10h
  }
}

void SetPal(char c, char r, char g, char b)
{
  asm
  {
      mov    dx,  03C8h
      mov    al,  c
      out    dx,  al
      inc    dx
      mov    al,  r
      out    dx,  al
      mov    al,  g
      out    dx,  al
      mov    al,  b
      out    dx,  al
  }
}

void BuildLightMap()
{
  double first_light, spot_light, r;

  for (int i = -R; i < R; i++)
    for (int j = -R; j < R; j++)
    {
      first_light = Vector(0, 0, 1) & Normalize(Vector(0, 0, light.org.z) - Vector(i, j, 0));
      r = sqrt(i * i + j * j);
      if (r < R - EDGE)
        spot_light = first_light;
      else
        if (r > R)
          spot_light = 0;
        else
          spot_light = first_light * (R - r) / EDGE;
      light_map[(j + R) * R_2 + i + R] = spot_light * 62 + random(2);
    }
}

void BuildTexture()
{
  for (int i = 0; i < TEX_SIZE; i++)
    for (int j = 0; j < TEX_SIZE; j++)
      if (i & (TEX_SIZE >> 1) ^ j & (TEX_SIZE >> 1))
        texture[(j << TEX_BITS) + i] = 0;
      else
        texture[(j << TEX_BITS) + i] = 1;
}

void AsmRender(int x1, int x2, int y1, int y2, int a, int b)
{
  asm
  {
      push   0A000h
      pop    es
      mov    di,  y1
      mov    ax,  di
      shl    di,  8
      shl    ax,  6
      add    di,  ax
      add    di,  x1
      mov    cx,  y2
      sub    cx,  y1
@@1:  push   cx
      mov    dx,  y2
      sub    dx,  cx
      mov    cx,  x2
      sub    cx,  x1
@@2:  mov    bx,  dx
      sub    bx,  b
      and    bl,  0FEh
      shl    bx,  R_BITS
      mov    ax,  x2
      sub    ax,  cx
      sub    ax,  a
      shr    ax,  1
      add    bx,  ax
      push   ds
      lds    si,  light_map
      add    si,  bx
      lodsb
      pop    ds
      mov    bx,  dx
      and    bx,  TEX_MASK
      shl    bx,  TEX_BITS
      push   ax
      mov    ax,  x2
      sub    ax,  cx
      and    ax,  TEX_MASK
      add    bx,  ax
      push   ds
      lds    si,  texture
      add    si,  bx
      lodsb
      pop    ds
      shl    al,  6
      xchg   bl,  al
      pop    ax
      add    bl,  al
      xchg   bl,  al
      stosb
      loop   @@2
      pop    cx
      sub    di,  x2
      add    di,  x1
      add    di,  320
      loop   @@1
  }
}

void Render()
{
  int x1 = light.org.x - R_2, a = x1;
  int x2 = light.org.x + R_2;
  int y1 = light.org.y - R_2, b = y1;
  int y2 = light.org.y + R_2;
  int light;

  if (x1 < 0) x1 = 0;
  if (x2 > MAX_X) x2 = MAX_X;
  if (y1 < 0) y1 = 0;
  if (y2 > MAX_Y) y2 = MAX_Y;

  AsmRender(x1, x2, y1, y2, a, b);
}

void main()
{
  double r, a;

  randomize();
  SetMode(0x13);

  for (int i = 0; i < 64; i++)
  {
    SetPal(i, i, i, i);
    SetPal(i + 64, 0, 0, i);
  }

  light_map = new char far[R * R * 4];
  texture = new char far[TEX_SIZE * TEX_SIZE];

  BuildLightMap();
  BuildTexture();

  do
  {
    Render();
    r = sin(a * M_PI / 90);
    light.org.x = cos(a * C) * r * R2 + MAX_X / 2;
    light.org.y = sin(a * C) * r * R3 + MAX_Y / 2;
    a++;
    if (a > 360)
      a -= 360;
  } while (!kbhit());

  delete texture;
  delete light_map;
  SetMode(0x03);
}
